# Copyright (C) 2022 Stefano Babic
#
# SPDX-License-Identifier: GPLv3
#
# This class manages to pack the SWU file
# it has methods to add files to the archive
import os
import stat


class CPIOException(Exception):
    pass


class SWUFile:
    def __init__(self, outfile):
        self.position = 0
        """
        :type outfile: FileIO of bytes
        """
        self.outfile = outfile
        self.position = 0

    def _rawwrite(self, data):
        """
        :type data: bytes
        """
        self.outfile.write(data)
        self.position += len(data)

    def _align(self):
        offset = -(self.position % -4)
        if offset:
            self._rawwrite(b"\x00" * offset)

    def cpiocrc(self, cpio_filename):
        crc = 0
        with open(cpio_filename, "rb") as rd:
            chunk = rd.read(65536)
            while chunk:
                for b in chunk:
                    crc += b
                chunk = rd.read(65536)
        crc = crc & 0xFFFFFFFF
        return crc

    def addartifacttoswu(self, cpio_filename):
        """
        :type cpio_filename: string
        """

        statres = os.stat(cpio_filename)
        size = statres.st_size

        # very common error, so check sooner
        if size > 0xFFFFFFFF:
            raise CPIOException(
                "Too big file size for this CPIO format", statres.st_size
            )

        if not stat.S_ISREG(statres.st_mode) or not size:
            raise CPIOException("File is not a regular file")

        if cpio_filename == b"TRAILER!!!":
            raise CPIOException("Attempt to pass reserved filename", cpio_filename)

        self.write_header(cpio_filename)

        self._align()
        with open(cpio_filename, "rb") as xxx:
            while size:
                chunk = xxx.read(65536)
                if not chunk:
                    break
                self._rawwrite(chunk)
                size -= len(chunk)
        if size:
            raise CPIOException("File was changed while reading", cpio_filename)

    def write_header(self, cpio_filename):
        if cpio_filename != "TRAILER!!!":
            statres = os.stat(cpio_filename)
            crc = self.cpiocrc(cpio_filename)
            base_filename = os.path.basename(cpio_filename)
            fields = [
                statres.st_ino,
                statres.st_mode,
                statres.st_uid,
                statres.st_gid,
                statres.st_nlink,
                0,  # mtime
                statres.st_size,
                os.major(statres.st_dev),
                os.minor(statres.st_dev),
                os.major(statres.st_rdev),
                os.minor(statres.st_rdev),
                # Yes, length including last zero byte
                len(base_filename) + 1,
                # New CRC Format
                #  The ``CRC'' format is mis-named, as it uses a simple checksum and not a
                #  cyclic redundancy check.
                crc,
            ]
        else:
            base_filename = cpio_filename
            fields = [0, 0, 0, 0, 16, 0, 0, 0, 0, 0, 0, len(base_filename) + 1, 0]

        self._align()
        self._rawwrite(b"070702")
        for i, value in enumerate(fields):
            # UNIX epoch overflow, negative timestamps and so on...
            if (value > 0xFFFFFFFF) or (value < 0):
                raise CPIOException("STOP: value out of range", i, value)
            s = f"{value:08X}"
            self.outfile.write(bytes(s, "ascii"))
            self.position += len(s)

        fn = bytes(base_filename, "ascii") + b"\x00"
        self._rawwrite(fn)

    def close(self):
        self.write_header("TRAILER!!!")
        offset = -(self.position % -512)
        if offset:
            self._rawwrite(b"\x00" * offset)
        self.position = 0



# SWUGenerator, SWU Package Generator for SWUpdate
#
# Copyright (C) 2022 Stefano Babic
#
# SPDX-License-Identifier: GPLv3

import argparse
import codecs
import logging
import os
import sys
import textwrap
from pathlib import Path

import libconf

from swugenerator import __about__, generator
from swugenerator.swu_sign import SWUSignCMS, SWUSignCustom, SWUSignPKCS11, SWUSignRSA


def extract_keys(keyfile):
    try:
        with open(keyfile, "r") as f:
            lines = f.readlines()
    except IOError:
        logging.fatal("Failed to open file with keys %s", keyfile)
        sys.exit(1)

    key, iv = None, None
    for line in lines:
        if "key" in line:
            key = line.split("=")[1].rstrip("\n")
        if "iv" in line:
            iv = line.split("=")[1].rstrip("\n")
    return key, iv


def main() -> None:
    """swugenerator main entry point."""

    # ArgumentParser {{{
    parser = argparse.ArgumentParser(
        prog=__about__.__title__,
        description=__about__.__summary__ + " " + __about__.__version__,
        fromfile_prefix_chars="@",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument(
        "-K",
        "--encryption-key-file",
        help="<key,iv> : AES Key to encrypt artifacts",
    )

    parser.add_argument(
        "-n",
        "--no-compress",
        action="store_const",
        const=True,
        default=False,
        help="Do not compress files",
    )

    parser.add_argument(
        "-e",
        "--no-encrypt",
        action="store_const",
        const=True,
        default=False,
        help="Do not encrypt files",
    )

    parser.add_argument(
        "-x",
        "--no-ivt",
        action="store_const",
        const=True,
        default=False,
        help="Do not generate IV when encrypting",
    )

    parser.add_argument(
        "-k",
        "--sign",
        help=textwrap.dedent(
            """\
            RSA key or certificate to sign the SWU
            One of :
            CMS,<private key>,<certificate used to sign>,<file with password if any>
            RSA,<private key>,<file with password if any>
            PKCS11,<pin>
            CUSTOM,<custom command> """
        ),
    )

    parser.add_argument(
        "-s",
        "--sw-description",
        required=True,
        help="sw-description template",
    )

    parser.add_argument(
        "-t",
        "--encrypt-swdesc",
        action="store_const",
        const=True,
        default=False,
        help="Encrypt sw-description",
    )

    parser.add_argument(
        "-a",
        "--artifactory",
        help="list of directories where artifacts are searched",
    )

    parser.add_argument(
        "-o",
        "--swu-file",
        required=True,
        help="SWU output file",
    )

    parser.add_argument(
        "-c",
        "--config",
        help="configuration file",
    )

    parser.add_argument(
        "-l",
        "--loglevel",
        help="set log level, default is WARNING",
    )
    parser.add_argument(
        "command", metavar="command", default=[], help="command to be executed (create)"
    )

    args = parser.parse_args()

    if args.loglevel:
        if args.loglevel == "DEBUG":
            logging.basicConfig(level=logging.DEBUG)
        if args.loglevel == "INFO":
            logging.basicConfig(level=logging.INFO)
        if args.loglevel == "ERROR":
            logging.basicConfig(level=logging.ERROR)
        if args.loglevel == "CRITICAL":
            logging.basicConfig(level=logging.CRITICAL)

    # Read configuration file if any
    config_vars = {}
    if args.config and args.config != "":
        logging.info("Reading configuration file %s", args.config)

        with codecs.open(args.config, "r", "utf-8") as f:
            config = libconf.load(f)
            for key, keydict in config.items():
                if key == "variables":
                    for varname, varvalue in keydict.items():
                        logging.debug("VAR = %s VAL = %s", varname, varvalue)
                        config_vars[varname] = varvalue
            f.close()

    # Signing
    sign_option = None
    if args.sign:
        sign_parms = args.sign.split(",")
        cmd = sign_parms[0]
        if cmd == "CMS":
            if len(sign_parms) < 3:
                logging.critical("CMS requires private key and certificate")
                sys.exit(1)
            # Format : CMS,<private key>,<certificate used to sign>,<file with password>
            if len(sign_parms) == 4:
                sign_option = SWUSignCMS(sign_parms[1], sign_parms[2], sign_parms[3])
            # Format : CMS,<private key>,<certificate used to sign>
            else:
                sign_option = SWUSignCMS(sign_parms[1], sign_parms[2], None)
        if cmd == "RSA":
            if len(sign_parms) < 2:
                logging.critical("RSA requires private key")
                sys.exit(1)
            # Format : RSA,<private key>,<file with password>
            if len(sign_parms) == 3:
                sign_option = SWUSignRSA(sign_parms[1], sign_parms[2])
            # Format : RSA,<private key>
            else:
                sign_option = SWUSignRSA(sign_parms[1], None)
        if cmd == "PKCS11":
            # Format : PKCS11,<pin>>
            if len(sign_parms) < 2:
                logging.critical("PKCS11 requires URI")
                sys.exit(1)
            sign_option = SWUSignPKCS11(sign_parms[1])
        if cmd == "CUSTOM":
            # Format : PKCS11,<custom command>>
            if len(sign_parms) < 2:
                logging.critical("PKCS11 requires URI")
                sys.exit(1)
            sign_option = SWUSignCustom(sign_parms[1])

    key = None
    iv = None
    if args.encryption_key_file:
        key, iv = extract_keys(args.encryption_key_file)

    artidirs = []
    artidirs.append(os.getcwd())
    if args.artifactory:
        dirs = args.artifactory.split(",")
        for directory in dirs:
            deploy = Path(directory).resolve()
            artidirs.append(deploy)

    if args.command == "create":
        swu = generator.SWUGenerator(
            args.sw_description,
            args.swu_file,
            config_vars,
            artidirs,
            sign_option,
            key,
            iv,
            args.encrypt_swdesc,
            args.no_compress,
            args.no_encrypt,
            args.no_ivt,
        )
        swu.process()
        swu.close()
    else:
        parser.error("no suitable command found: (create)")


if __name__ == "__main__":
    main()



# Copyright (C) 2022 Stefano Babic
#
# SPDX-License-Identifier: GPLv3
import copy
import logging
import subprocess
import sys


class SWUSign:
    def __init__(self):
        self.type = None
        self.key = None
        self.cert = None
        self.cmd = None
        self.passin = None
        self.signcmd = []

    def get_passwd_file_args(self):
        passwd_args = []
        if self.passin:
            passwd_args = ["-passin", f"file:{self.passin}"]
        return passwd_args

    def set_password_file(self, passin):
        self.passin = passin

    def sign(self):
        try:
            subprocess.run(" ".join(self.signcmd), shell=True, check=True, text=True)
        except subprocess.CalledProcessError:
            logging.critical(
                "SWU cannot be signed, signing command was %s", self.signcmd
            )
            sys.exit(1)


class SWUSignCMS(SWUSign):
    def __init__(self, key, cert, passin):
        super().__init__()
        self.type = "CMS"
        self.key = key
        self.cert = cert
        self.passin = passin

    def prepare_cmd(self, sw_desc_in, sw_desc_sig):
        self.signcmd = [
            "openssl",
            "cms",
            "-sign",
            "-in",
            sw_desc_in,
            "-out",
            sw_desc_sig,
            "-signer",
            self.cert,
        ]
        self.signcmd += [
            "-inkey",
            self.key,
            "-outform",
            "DER",
            "-nosmimecap",
            "-binary",
        ]
        self.signcmd += self.get_passwd_file_args()


class SWUSignRSA(SWUSign):
    def __init__(self, key, passin):
        super().__init__()
        self.type = "RSA"
        self.key = key
        self.passin = passin

    def prepare_cmd(self, sw_desc_in, sw_desc_sig):
        self.signcmd = (
            ["openssl", "dgst", "-sha256", "-sign", self.key]
            + self.get_passwd_file_args()
            + ["-out", sw_desc_sig, sw_desc_in]
        )


class SWUSignCustom(SWUSign):
    def __init__(self, cmd):
        super().__init__()
        self.type = "CUSTOM"
        self.custom = cmd.split()

    def prepare_cmd(self, *_):
        self.signcmd = copy.deepcopy(self.custom)


# Note: tested with Nitrokey HSM
class SWUSignPKCS11(SWUSign):
    def __init__(self, pin):
        super().__init__()
        self.type = "PKCS11"
        self.custom = ["--pin"]
        self.custom.append(pin)

    def prepare_cmd(self, sw_desc_in, sw_desc_sig):
        self.signcmd = [
            "pkcs11-tool",
            "-s",
            "-m ",
            "SHA256-RSA-PKCS",
            "-i",
            sw_desc_in,
            "-o",
            sw_desc_sig,
        ] + self.custom
